# -------------------------------------------------------------
# NDN Hydra Query Client
# -------------------------------------------------------------
#  @Project: NDN Hydra
#  @Authors: Please check AUTHORS.rst
#  @Source-Code:   https://github.com/tntech-ngin/ndn-hydra
#  @Documentation: https://ndn-hydra.readthedocs.io
#  @Pip-Library:   https://pypi.org/project/ndn-hydra
# -------------------------------------------------------------

import logging
from ndn.app import NDNApp
from ndn.encoding import FormalName, Component, Name, ContentType
from ndn.types import InterestNack, InterestTimeout, InterestCanceled, ValidationFailure
from ndn_hydra.repo.protocol.base_models import FileList, File


class HydraQueryClient(object):
    def __init__(self, app: NDNApp, client_prefix: FormalName, repo_prefix: FormalName) -> None:
        """
        This client queries a node within the remote repo.
        :param app: NDNApp.
        :param client_prefix: NonStrictName. Routable name to client.
        :param repo_prefix: NonStrictName. Routable name to remote repo.
        """
        self.app = app
        self.client_prefix = client_prefix
        self.repo_prefix = repo_prefix

    async def send_query(self, query: Name, node_name: str = None) -> None:
        """
        Form a certain query and request that info from a node.
        """
        if not node_name:
            named_query = self.repo_prefix + [Component.from_str("query")]
        else:
            named_query = self.repo_prefix + [Component.from_str("node")] + [Component.from_str(node_name)] + [Component.from_str("query")]

        try:
            data_name, meta_info, content = await self.app.express_interest(named_query, Name.to_bytes(query), can_be_prefix=True, must_be_fresh=True, lifetime=3000)

            if meta_info.content_type == ContentType.NACK:
                print("Distributed Repo does not know that query.")
            else:
                querytype = Component.to_str(query[0])
                if querytype == "nodes":
                    print(f'List of All Node Names')
                    print(f'{bytes(content).decode().split()}')
                elif querytype == "exnodes":
                    print(f'List of All Expired Node Names')
                    print(f'{bytes(content).decode().split()}')
                elif querytype == "files":
                    filelist = FileList.parse(content)
                    counter = 1
                    if filelist.list:
                        print(f'\nList of All Files')
                        for file in filelist.list:
                            print(f'File {counter} meta-info')
                            print(f'\tfile_name: {Name.to_str(file.file_name)}')
                            print(f'\tsize: {file.size} | packets: {file.packets} | packet_size: {file.packet_size}')
                            counter = counter + 1
                    else:
                        print(f'No files inserted in the remote repo.')
                elif querytype == "file":
                    if content:
                        file = File.parse(content)
                        print(f'File meta-info:')
                        print(f'\tfile_name: {Name.to_str(file.file_name)}')
                        print(f'\tsize: {file.size} | packets: {file.packets} | packet_size: {file.packet_size}')
                        return Name.to_str(file.file_name)
                elif querytype == "prefix":
                    filelist = FileList.parse(content)
                    counter = 1
                    if filelist.list:
                        print(f'List of All Files with prefix {Name.to_str(query[1:])}')
                        for file in filelist.list:
                            print(f'File {counter} meta-info')
                            print(f'\tfile_name: {Name.to_str(file.file_name)}')
                            print(f'\tsize: {file.size} | packets: {file.packets} | packet_size: {file.packet_size}')
                            counter = counter + 1
                    else:
                        print(f'No files inserted in the remote repo with prefix {Name.to_str(query[1:])}.')
                else:
                    print("Client does not know that query.")
        except (InterestNack, InterestTimeout, InterestCanceled, ValidationFailure) as e:
            print("Query command received no data packet back")


import sqlite3
from typing import List

from ndn.storage import SqliteStorage
from ndn.encoding import Name, NonStrictName


# DataStorage adds batch remove packets method
# to the SqliteStorage
class DataStorage(SqliteStorage):
    def __init__(self, db_path: str, write_period: int = 10, initialize: bool = True):
        super().__init__(db_path, write_period, initialize)
        # self.conn.execute('PRAGMA journal_mode=wal')  # Enable WAL mode

    def remove_packets(self, names: List[NonStrictName]) -> int:
        if self.initialized is not True:
            raise self.UninitializedError("The storage is not initialized.")

        keys = [self._get_name_bytes_wo_tl(Name.normalize(name)) for name in names]
        if not keys:
            return 0

        total_deleted = 0
        batch_size = 1000
        try:
            self.conn.execute('BEGIN TRANSACTION')

            for i in range(0, len(keys), batch_size):
                batch_keys = keys[i:i + batch_size]
                placeholders = ', '.join('?' * len(batch_keys))
                query = f'DELETE FROM data WHERE key IN ({placeholders})'

                c = self.conn.cursor()
                c.execute(query, batch_keys)
                total_deleted += c.rowcount

            self.conn.commit()
        except sqlite3.Error as e:
            self.conn.rollback()
            print(f"SQLite error: {e}")

        finally:
            if c: c.close()

        try:
            self.conn.execute('VACUUM')
        except sqlite3.Error as e:
            print(f"SQLite error during VACUUM: {e}")

        return total_deleted



# -------------------------------------------------------------
# NDN Hydra MainLoop
# -------------------------------------------------------------
#  @Project: NDN Hydra
#  @Authors: Please check AUTHORS.rst
#  @Source-Code:   https://github.com/justincpresley/ndn-hydra
#  @Documentation: https://ndn-hydra.readthedocs.io
#  @Pip-Library:   https://pypi.org/project/ndn-hydra
# -------------------------------------------------------------

import asyncio as aio
import logging
import secrets
import time
import random
from typing import Dict, List
from ndn.app import NDNApp
from ndn.encoding import Name, Component
from ndn.types import InterestNack, InterestTimeout
from ndn.svs import SVSync
from ndn.storage import Storage, SqliteStorage
from ndn_hydra.repo.modules import *
from ndn_hydra.repo.group_messages import *
from ndn_hydra.repo.modules.file_fetcher import FileFetcher
from ndn_hydra.repo.utils.garbage_collector import collect_db_garbage
from ndn_hydra.repo.utils.concurrent_fetcher import concurrent_fetcher
from ndn_hydra.repo.modules.favor_calculator import FavorCalculator
from ndn_hydra.repo.modules.read_remaining_space import get_remaining_space


class MainLoop:
    def __init__(self, app: NDNApp, config: Dict, global_view: GlobalView, data_storage: Storage, svs_storage: Storage, file_fetcher: FileFetcher):
        self.app = app
        self.config = config
        self.global_view = global_view
        self.data_storage = data_storage
        self.svs_storage = svs_storage
        self.file_fetcher = file_fetcher
        self.file_fetcher.store_func = self.store
        self.svs = None
        self.logger = logging.getLogger('ndn')
        self.node_name = self.config['node_name']
        self.tracker = HeartbeatTracker(self.node_name, global_view, config['loop_period'], config['heartbeat_rate'], config['tracker_rate'], config['beats_to_fail'], config['beats_to_renew'])
        self.last_garbage_collect_t = time.time()  # time in seconds
        self.favor = 0

    async def start(self):
        self.svs = SVSync(self.app,
                          Name.normalize(self.config['repo_prefix'] + "/group"),
                          Name.normalize(self.node_name),
                          self.svs_missing_callback,
                          storage=self.svs_storage)
        await aio.sleep(5)
        while True:
            await aio.sleep(self.config['loop_period'] / 1000.0)
            self.periodic()

    def periodic(self):
        self.tracker.detect()
        if self.tracker.beat():
            self.send_heartbeat()
            self.tracker.reset(self.node_name)
        self.backup_list_check()
        self.claim()
        self.check_garbage()

    def svs_missing_callback(self, missing_list):
        aio.ensure_future(self.on_missing_svs_messages(missing_list))

    async def on_missing_svs_messages(self, missing_list):
        # if missing list is greater than 100 messages, bootstrap
        for i in missing_list:
            if i.nid == self.config["node_name"]:
                self.tracker.restart(self.config["node_name"])
                # bootstrap
                continue
            while i.lowSeqno <= i.highSeqno:
                message_bytes = await self.svs.fetchData(Name.from_str(i.nid), i.lowSeqno)
                if message_bytes is None:
                    continue
                message = Message.specify(i.nid, i.lowSeqno, message_bytes)
                self.tracker.reset(i.nid)
                aio.ensure_future(message.apply(self.global_view, self.data_storage, self.fetch_file, self.svs, self.config))
                i.lowSeqno = i.lowSeqno + 1

    def send_heartbeat(self):
        heartbeat_message = HeartbeatMessageTlv()
        heartbeat_message.node_name = self.config['node_name'].encode()

        node_path = "/".join(self.config['data_storage_path'].split("/")[:-1])
        remaining_space = get_remaining_space(node_path)

        logging.debug(f"\n[MAIN LOOP][SEND_HEARTBEAT] "
                          f"\n\tRemaining space for node {self.config['node_name']} is: {remaining_space}")

        # Create FavorParameter and fill its fields
        favor_parameters = FavorParameters()
        favor_parameters.rtt = str(self.config['rtt'])
        favor_parameters.num_users = str(self.config['num_users'])
        favor_parameters.bandwidth = str(self.config['bandwidth'])
        favor_parameters.network_cost = str(self.config['network_cost'])
        favor_parameters.storage_cost = str(self.config['storage_cost'])
        favor_parameters.remaining_storage = str(remaining_space)
        favor_parameters.rw_speed = str(self.config['rw_speed'])

        heartbeat_message.favor_parameters = favor_parameters

        # Create FavorWeights and set its fields
        favor_weights = FavorWeights()
        favor_weights.remaining_storage = '0.14'
        favor_weights.bandwidth = '0'
        favor_weights.rw_speed = '0'

        # Assign the encoded FavorWeights
        heartbeat_message.favor_weights = favor_weights

        self_favor = FavorCalculator.calculate_favor(
            {
                'remaining_storage': remaining_space,
                'bandwidth': self.config['bandwidth'],
                'rw_speed': self.config['rw_speed']
            },
            {
                'remaining_storage': 0.14,
                'bandwidth': 0,
                'rw_speed': 0
            })

        message_to_send = Message()
        message_to_send.type = MessageTypes.HEARTBEAT
        message_to_send.value = heartbeat_message.encode()

        try:
            next_state_vector = self.svs.getCore().getStateTable().getSeqno(Name.to_str(Name.from_str(self.config['node_name']))) + 1
        except TypeError:
            next_state_vector = 0

        # Update favor for this node in global_view
        self.global_view.update_node(self.config['node_name'], self_favor, next_state_vector)
        self.svs.publishData(message_to_send.encode())

        self.logger.debug(f"\n[MAIN LOOP][SEND_HEARTBEAT] "
                          f"\n\tNode {self.config['node_name']} favor is: {self_favor}")

        self.logger.debug(f"\n[MAIN LOOP][SEND_HEARTBEAT] "
                          f"\n\tGlobal view for node {self.config['node_name']} is:"
                          f"\n\n----------/----------/----------/----------"
                          f"\n{self.global_view}"
                          f"----------/----------/----------/----------\n")

    def backup_list_check(self):
        underreplicated_files = self.global_view.get_underreplicated_files()
        for underreplicated_file in underreplicated_files:
            deficit = underreplicated_file['desired_copies'] - len(underreplicated_file['stores'])
            for backuped_by in underreplicated_file['backups']:
                if (backuped_by['node_name'] == self.config['node_name']) and (backuped_by['rank'] < deficit):
                    self.fetch_file(underreplicated_file['file_name'], underreplicated_file['packets'], underreplicated_file['packet_size'], underreplicated_file['fetch_path'])

    def claim(self):
        # TODO: possibility based on # active sessions and period
        if random.random() < 0.618:
            return
        backupable_files = self.global_view.get_backupable_files()
        for backupable_file in backupable_files:
            if random.random() < 0.618:
                continue
            # print(json.dumps(backupable_insertion['stores']))
            # print(json.dumps(backupable_insertion['backups']))
            already_in = False
            for stored_by in backupable_file['stores']:
                if stored_by == self.config['node_name']:
                    already_in = True
                    break
            for backuped_by in backupable_file['backups']:
                if backuped_by['node_name'] == self.config['node_name']:
                    already_in = True
                    break
            if already_in is True:
                continue
            if len(backupable_file['backups']) == 0 and len(backupable_file['stores']) == 0:
                continue
            authorizer = None
            if len(backupable_file['backups']) == 0:
                authorizer = {
                    'node_name': backupable_file['stores'][-1],
                    'rank': -1,
                    'nonce': backupable_file['file_name']
                }
            else:
                authorizer = backupable_file['backups'][-1]
            # generate claim (request) msg and send
            # claim tlv
            claim_message = ClaimMessageTlv()
            claim_message.node_name = self.config['node_name'].encode()
            claim_message.favor = str(self.global_view.get_node(self.config['node_name'])['favor']).encode()
            claim_message.file_name = Name.from_str(backupable_file['file_name'])
            claim_message.type = ClaimTypes.REQUEST
            claim_message.claimer_node_name = self.config['node_name'].encode()
            claim_message.claimer_nonce = secrets.token_hex(4).encode()
            claim_message.authorizer_node_name = authorizer['node_name'].encode()
            claim_message.authorizer_nonce = authorizer['nonce'].encode()

            # claim msg
            message = Message()
            message.type = MessageTypes.CLAIM
            message.value = claim_message.encode()
            self.svs.publishData(message.encode())
            self.logger.info(f"\n[MSG][CLAIM.R]* "
                             f"\n\tNode name={self.config['node_name']};"
                             f"\n\tFile name={backupable_file['file_name']}")

    def store(self, file_name: str):
        store_message = StoreMessageTlv()
        store_message.node_name = self.config['node_name'].encode()

        store_message.favor = str(self.global_view.get_node(self.config['node_name'])['favor']).encode()
        store_message.file_name = Name.from_str(file_name)
        message = Message()
        message.type = MessageTypes.STORE
        message.value = store_message.encode()

        self.global_view.store_file(file_name, self.config['node_name'])
        self.svs.publishData(message.encode())
        self.logger.info(f"\n[MSG][STORE]* "
                         f"\n\tNode name={self.config['node_name']};"
                         f"\n\tfile={file_name}")

    def fetch_file(self, file_name: str, packets: int, packet_size: int, fetch_path: str):
        self.file_fetcher.fetch_file_from_client(file_name, packets, packet_size, fetch_path)

    def check_garbage(self):
        """
        Checks for database garbage daily.
        """
        current_time = time.time()
        hours_since_last_collection = (current_time - self.last_garbage_collect_t) / (60 * 60)
        if hours_since_last_collection >= 24:
            collect_db_garbage(self.global_view,
                               self.data_storage,
                               self.svs,
                               self.config,
                               self.logger)
            self.last_garbage_collect_t = time.time()



# -------------------------------------------------------------
#  @Project: NDN Hydra
#  @Authors: Please check AUTHORS.rst
#  @Source-Code:   https://github.com/tntech-ngin/ndn-hydra
#  @Documentation: https://ndn-hydra.readthedocs.io
#  @Pip-Library:   https://pypi.org/project/ndn-hydra
# -------------------------------------------------------------

from .insert import HydraInsertClient
from .fetch import HydraFetchClient
from .delete import HydraDeleteClient
from .query import HydraQueryClient



# -------------------------------------------------------------
# NDN Hydra Main
# -------------------------------------------------------------
#  @Project: NDN Hydra
#  @Authors: Please check AUTHORS.rst
#  @Source-Code:   https://github.com/tntech-ngin/ndn-hydra
#  @Documentation: https://ndn-hydra.readthedocs.io
#  @Pip-Library:   https://pypi.org/project/ndn-hydra
# -------------------------------------------------------------

import logging
from argparse import ArgumentParser
from typing import Dict
from threading import Thread
import pkg_resources
from ndn.app import NDNApp
from ndn.encoding import Name
from ndn.storage import SqliteStorage
import sys, os
from ndn.svs import SVSyncLogger
from ndn_hydra.repo import *
from ndn_hydra.repo.modules.file_fetcher import FileFetcher
from ndn_hydra.repo.modules.data_storage import DataStorage
from ndn_hydra.repo.modules.read_config import read_config_file


def process_cmd_opts():
    def interpret_version(args) -> None:
        if args.version and (len(sys.argv) - 1 < 2):
            try:
                print("ndn-hydra " + pkg_resources.require("ndn-hydra")[0].version)
            except pkg_resources.DistributionNotFound:
                print("ndn-hydra source, undetermined")
            sys.exit(0)

    def interpret_help(args) -> None:
        if args.help:
            if len(sys.argv) - 1 < 2:
                print("* Basic initialization:")
                print("    ndn-hydra-repo [-h] [-v] -rp REPO_PREFIX -n NODE_NAME")
                print("")
                print("    ndn-hydra-repo: hosting a node for hydra, the NDN distributed repo.")
                print("* Examples:")
                print("    ('python3 ./examples/repo.py' instead of 'ndn-hydra-repo' if from source.)")
                print("")
                print("* Informative arguments:")
                print("    -h, --help                        |   shows this help message and exits.")
                print("    -v, --version                     |   shows the current version and exits.")
                print("")
                print("* Optional arguments:")
                print("    -rp, --repoprefix REPO_PREFIX     |   repo (group) prefix. Example: \"/hydra\"")
                print("    -n,  --nodename NODE_NAME         |   node name. Example: \"node01\"")
                print("    -d,  --debugger                   |   enable debugging mode")
                print("    -cr, --critical                   |   enable critical logging level")
                print("")
                print("* Default configuration:")
                print("    The default configuration is in the config.yaml file in the repo folder: ")
                print("    ndn_hydra/repo/config.yaml")
                print("    You can update these parameters there and test again: ")
                print("")
                print("  ** Parameters **")
                print("    Repo prefix  | Prefix for all nodes in the repo. Example: \"/hydra\"")
                print("    Node name    | Name of the node. Example: \"node01\"")
                print("")
                print("  ** Paths **")
                print("    Base         | Base path for all files. Example: \"/home/user/.ndn\"")
                print("    Data storage | Path for data storage. Example: \"/home/user/.ndn/hydra/node01/data.db\"")
                print(
                    "    Global view  | Path for global view. Example: \"/home/user/.ndn/hydra/node01/global_view.db\"")
                print("    Svs storage  | Path for svs storage. Example: \"/home/user/.ndn/hydra/node01/svs.db\"")
                print("    Logging      | Path for logging. Example: \"/home/user/.ndn/hydra/node01/session.log\"")
                print("")
                print("Thank you for using hydra.")
            sys.exit(0)

    def process_name(input_string: str):
        if input_string[-1] == "/":
            input_string = input_string[:-1]
        if input_string[0] != "/":
            input_string = "/" + input_string
        return input_string

    def parse_cli_args():
        # Command Line Parser
        parser = ArgumentParser(prog="ndn-hydra-repo", add_help=False, allow_abbrev=False)

        # Adding all Command Line Arguments
        parser.add_argument("-h", "--help", action="store_true", dest="help", default=False, required=False)
        parser.add_argument("-v", "--version", action="store_true", dest="version", default=False, required=False)
        parser.add_argument("-rp", "--repoprefix", action="store", dest="repo_prefix", default=False, required=False)
        parser.add_argument("-n", "--nodename", action="store", dest="node_name", default=False, required=False)
        parser.add_argument("-d", "--debugger", action="store_true", dest="debugger", default=False, required=False)
        parser.add_argument("-cr", "--critical", action="store_true", dest="critical", default=False, required=False)

        # Getting all Arguments
        cli_args = parser.parse_args()

        # Interpret Informational Arguments
        interpret_version(cli_args)
        interpret_help(cli_args)

        return cli_args

    def create_config():
        cli_args = parse_cli_args()
        # Get values from YAML file
        default_config_file = read_config_file()

        default_repo_prefix = default_config_file['default_config']['repo_prefix']
        default_node_name = default_config_file['default_config']['node_name']

        config_data = {
            "repo_prefix": default_repo_prefix,
            "node_name": default_node_name,
            "loop_period": default_config_file['default_config']['timers']['loop_period'],
            "heartbeat_rate": default_config_file['default_config']['timers']['heartbeat_rate'],
            "tracker_rate": default_config_file['default_config']['timers']['tracker_rate'],
            "beats_to_fail": default_config_file['default_config']['timers']['beats_to_fail'],
            "beats_to_renew": default_config_file['default_config']['timers']['beats_to_renew'],
            "replication_degree": default_config_file['default_config']['timers']['replication_degree'],
            "file_expiration": default_config_file['default_config']['timers']['file_expiration'],
            "rtt": default_config_file['default_config']['favor']['rtt'],
            "num_users": default_config_file['default_config']['favor']['num_users'],
            "bandwidth": default_config_file['default_config']['favor']['bandwidth'],
            "network_cost": default_config_file['default_config']['favor']['network_cost'],
            "storage_cost": default_config_file['default_config']['favor']['storage_cost'],
            "remaining_storage": default_config_file['default_config']['favor']['remaining_storage'],
            "rw_speed": default_config_file['default_config']['favor']['rw_speed'],
            "logger_level": default_config_file['default_config']['logger_level'],
        }

        if cli_args.repo_prefix is not False:
            config_data["repo_prefix"] = process_name(cli_args.repo_prefix)
        if cli_args.node_name is not False:
            config_data["node_name"] = process_name(cli_args.node_name)
        if cli_args.debugger is not False:
            config_data["logger_level"] = "DEBUG"
        if cli_args.critical is not False:
            config_data["logger_level"] = "CRITICAL"

        workpath = "{home}/.ndn/repo{repo_prefix}/{node_name}".format(
            home=os.path.expanduser("~"),
            repo_prefix=config_data["repo_prefix"],
            node_name=config_data["node_name"])
        config_data["logging_path"] = "{workpath}/session.log".format(workpath=workpath)
        config_data["data_storage_path"] = "{workpath}/data.db".format(workpath=workpath)
        config_data["global_view_path"] = "{workpath}/global_view.db".format(workpath=workpath)
        config_data["svs_storage_path"] = "{workpath}/svs.db".format(workpath=workpath)
        return config_data

    configuration = create_config()
    return configuration


async def listen(repo_prefix: Name, pb: PubSub, insert_handle: InsertCommandHandle, delete_handle: DeleteCommandHandle):
    # pubsub
    pb.set_publisher_prefix(repo_prefix)
    await pb.wait_for_ready()
    # protocol handle
    await insert_handle.listen(repo_prefix)
    await delete_handle.listen(repo_prefix)


class HydraNodeThread(Thread):
    def __init__(self, config: Dict):
        Thread.__init__(self)
        self.config = config

    def run(self) -> None:
        if 'logging_path' not in self.config or self.config['logging_path'] is None:
            raise ValueError("The 'logging_path' was not set in the configuration.")

        logging_dir = os.path.dirname(self.config['logging_path'])

        if logging_dir and not os.path.exists(logging_dir):
            try:
                os.makedirs(logging_dir)
            except PermissionError:
                raise PermissionError(f"Could not create directory: {logging_dir}")
            except FileExistsError:
                pass

        # logging

        log_level = getattr(logging, self.config['logger_level'].upper(), logging.INFO)

        logging.basicConfig(level=log_level,
                            format='%(levelname)-8s  %(message)s',
                            filename=self.config['logging_path'],
                            filemode='w')
        console = logging.StreamHandler()
        console.setLevel(log_level)
        logging.getLogger().addHandler(console)

        SVSyncLogger.config(False, None, logging.CRITICAL)

        logging.getLogger('ndn').setLevel(logging.WARNING)

        # NDN
        app = NDNApp()

        # Post-start
        async def start_main_loop():
            # databases
            data_storage = DataStorage(self.config['data_storage_path'])
            global_view = GlobalView(self.config['global_view_path'])
            svs_storage = SqliteStorage(self.config['svs_storage_path'])
            pb = PubSub(app)

            # file fetcher module
            file_fetcher = FileFetcher(app, global_view, data_storage, self.config)

            # main_loop (svs)
            main_loop = MainLoop(app, self.config, global_view, data_storage, svs_storage, file_fetcher)

            # handles (reads, commands & queries)
            read_handle = ReadHandle(app, data_storage, global_view, main_loop, self.config)
            insert_handle = InsertCommandHandle(app, data_storage, pb, self.config, main_loop, global_view)
            delete_handle = DeleteCommandHandle(app, data_storage, pb, self.config, main_loop, global_view)
            query_handle = QueryHandle(app, global_view, self.config)

            await listen(Name.normalize(self.config['repo_prefix']), pb, insert_handle, delete_handle)
            await main_loop.start()

        # start listening
        try:
            app.run_forever(after_start=start_main_loop())
        except (FileNotFoundError, ConnectionRefusedError):
            print('Error: could not connect to NFD.')
            sys.exit()


def main() -> int:
    config_args = process_cmd_opts()

    try:
        HydraNodeThread(config_args).start()
        return 0

    except Exception as e:
        logging.warning(f"\nAn error occurred running the main thread: {e}")
        return 1


if __name__ == "__main__":
    sys.exit(main())
